@@ -12,10 +12,12 @@ from typing import (  # noqa: F401
     Any, Callable, Dict, List, Optional, Tuple, Union, cast)
 
 import click
-from click_spinner import spinner
+from click_spinner import Spinner, spinner
+import jwt
 from pick import pick
 from pyvss.const import __version__ as pyvss_version
 from pyvss.enums import RequestStatus
+from pyvss.exceptions import VssError
 from pyvss.manager import VssManager
 from ruamel.yaml import YAML
 
@@ -23,11 +25,12 @@ import vss_cli.const as const
 from vss_cli.data_types import ConfigEndpoint, ConfigFile, ConfigFileGeneral
 from vss_cli.exceptions import VssCliError
 from vss_cli.helper import (
-    debug_requests_on, format_output, get_hostname_from_url)
+    bytes_to_str, debug_requests_on, format_output, get_hostname_from_url)
 from vss_cli.utils.emoji import EMOJI_UNICODE
 from vss_cli.utils.threading import WorkerQueue
 from vss_cli.validators import (
-    validate_email, validate_phone_number, validate_uuid, validate_vm_moref)
+    validate_email, validate_json_file_or_type, validate_phone_number,
+    validate_uuid, validate_vm_moref)
 import vss_cli.yaml as yaml
 
 _LOGGING = logging.getLogger(__name__)
@@ -50,9 +53,11 @@ class Configuration(VssManager):
         self.endpoint_name = const.DEFAULT_ENDPOINT_NAME
         # end of endpoint settings
         self.history = const.DEFAULT_HISTORY  # type: str
-        self.webdav_server = const.DEFAULT_WEBDAV_SERVER  # type: str
+        self.s3_server = None  # type: Optional[str]
+        self._s3_server = const.DEFAULT_S3_SERVER  # type: str
         self.username = None  # type: Optional[str]
         self.password = None  # type: Optional[str]
+        self.totp = None  # type: Optional[str]
         self.token = None  # type: Optional[str]
         self.timeout = None  # type: Optional[int]
         self._debug = False  # type: Optional[bool]
@@ -107,6 +112,7 @@ class Configuration(VssManager):
         self.base_endpoint = value
         self.api_endpoint = f'{value}/v2'
         self.token_endpoint = f'{value}/auth/request-token'
+        self.tf_endpoint = f'{value}/tf'
         if value:
             self.endpoint_name = get_hostname_from_url(
                 value, const.DEFAULT_HOST_REGEX
@@ -120,9 +126,15 @@ class Configuration(VssManager):
                 setattr(self, setting, default)
         _LOGGING.debug(self)
 
-    def get_token(self, user: str = '', password: str = '') -> str:
+    def get_token(
+        self,
+        user: Optional[str] = '',
+        password: Optional[str] = '',
+        otp: Optional[str] = None,
+    ) -> str:
         """Generate token and returns value."""
-        self.api_token = super().get_token(user, password)
+        self.api_token = super().get_token(user, password, otp)
+        _LOGGING.debug(f'Token generated successfully: {self.api_token}')
         return self.api_token
 
     def update_endpoints(self, endpoint: str = '') -> None:
@@ -130,6 +142,7 @@ class Configuration(VssManager):
         self.base_endpoint = endpoint
         self.api_endpoint = f'{endpoint}/v2'
         self.token_endpoint = f'{endpoint}/auth/request-token'
+        self.tf_endpoint = f'{endpoint}/tf'
 
     def echo(self, msg: str, *args: Optional[Any]) -> None:
         """Put content message to stdout."""
@@ -173,12 +186,14 @@ class Configuration(VssManager):
             "access_token": 'yes' if self.token is not None else 'no',
             "user": 'yes' if self.username is not None else 'no',
             "user_password": 'yes' if self.password is not None else 'no',
+            "tf_enabled": 'yes' if self.tf_enabled is not None else 'no',
             "output": self.output,
             "timeout": self.timeout,
             "debug": self.debug,
             "verbose": self.verbose,
             "wait": self.wait,
             "dry_run": self.dry_run,
+            "s3_server": self.s3_server,
         }
 
         return f"<Configuration({view})"
@@ -250,9 +265,13 @@ class Configuration(VssManager):
                 credentials_decoded = b64decode(auth_enc)
                 # get user/pass
                 username, password = credentials_decoded.split(b':')
-            return config_endpoint[0], username, password
+            return (
+                config_endpoint[0],
+                bytes_to_str(username),
+                bytes_to_str(password),
+            )
         else:
-            return None, username, password
+            return None, bytes_to_str(username), bytes_to_str(password)
 
     def load_config_file(
         self, config: Union[Path, str] = None
@@ -282,7 +301,7 @@ class Configuration(VssManager):
             )
 
     def load_config(
-        self, validate: bool = True
+        self, validate: bool = True, spinner_cls: Optional[Spinner] = None
     ) -> Optional[Tuple[str, str, str]]:
         """Load configuration and validate.
 
@@ -309,8 +328,7 @@ class Configuration(VssManager):
                         'Consider running command "configure mk" to save your '
                         'credentials.'
                     )
-                    self.get_token(self.username, self.password)
-                    _LOGGING.debug(f'Token generated {self.api_token}')
+                    self.get_token(self.username, self.password, self.totp)
                     return self.username, self.password, self.api_token
                 else:
                     raise VssCliError(
@@ -407,7 +425,11 @@ class Configuration(VssManager):
                             )
                         # set config_path data
                         self.set_credentials(
-                            usr, pwd, ep.token, ep.url, ep.name,
+                            usr,
+                            pwd,
+                            ep.token,
+                            ep.url,
+                            ep.name,
                         )
                         if validate:
                             # last check cred
@@ -430,17 +452,9 @@ class Configuration(VssManager):
                             except Exception as ex:
                                 self.vlog(str(ex))
                                 _LOGGING.debug('Generating a new token')
-                                try:
-                                    self.api_token = self.get_token(
-                                        self.username, self.password
-                                    )
-                                    _LOGGING.debug(
-                                        'Token generated successfully'
-                                    )
-                                except Exception as ex:
-                                    _LOGGING.warning(
-                                        f'Could not generate new token: {ex}'
-                                    )
+                                self.api_token = self._get_token_with_mfa(
+                                    spinner_cls=spinner_cls
+                                )
                                 endpoint = self._create_endpoint_config(
                                     token=self.api_token
                                 )
@@ -538,13 +552,53 @@ class Configuration(VssManager):
         except Exception as ex:
             _LOGGING.error(f'Could not check for messages: {ex}')
 
+    def _get_token_with_mfa(
+        self,
+        token: Optional[str] = None,
+        spinner_cls: Optional[Spinner] = None,
+    ):
+        """Get token with MFA."""
+        try:
+            token = token or self.get_token(
+                self.username, self.password, self.totp
+            )
+        except VssError as ex:
+            if (
+                500 > ex.http_code > 399
+                and 'InvalidParameterValue: otp' in ex.message
+            ):
+                # try MFA Auth
+                try:
+                    _LOGGING.debug('Requesting a new timed one-time password')
+                    _ = self.request_totp(self.username, self.password)
+                except Exception as exc:
+                    _LOGGING.warning(f'Requesting totp: {exc}')
+                    pass
+                if spinner_cls is not None:
+                    spinner_cls.stop()
+                self.totp = click.prompt(
+                    'MFA enabled. Provide Timed One-Time Password'
+                )
+                if spinner_cls is not None:
+                    spinner_cls.start()
+                try:
+                    token = token or self.get_token(
+                        self.username, self.password, self.totp
+                    )
+                except Exception as exc:
+                    _LOGGING.warning(f'Could not generate new token: {exc}')
+                    raise exc
+            else:
+                raise ex
+        return token
+
     def _create_endpoint_config(self, token: str = None) -> ConfigEndpoint:
         """Create endpoint configuration for a given token.
 
         Token might be ``None`` and will generate a new one
         using ``username`` and ``password``.
         """
-        token = token or self.get_token(self.username, self.password)
+        token = self._get_token_with_mfa(token=token)
         # encode or save
         username = (
             self.username
@@ -558,13 +612,20 @@ class Configuration(VssManager):
         )
         credentials = b':'.join([username, password])
         auth = b64encode(credentials).strip().decode('utf-8')
+        # decode jwt to verify if otp is enabled.
+        payload = jwt.decode(
+            self.api_token, options=dict(verify_signature=False)
+        )
         endpoint_cfg = {
             'url': self.base_endpoint,
             'name': self.endpoint_name,
             'auth': auth,
-            'token': token,
+            'token': self.api_token,
+            'tf_enabled': payload.get('otp', False),
         }
-        return ConfigEndpoint.from_json(json.dumps(endpoint_cfg))
+        ep_cfg = ConfigEndpoint.from_json(json.dumps(endpoint_cfg))
+        _LOGGING.debug(f'Configuration endpoint created: {ep_cfg}')
+        return ep_cfg
 
     def load_config_template(self) -> ConfigFile:
         """Load configuration from template."""
@@ -633,7 +694,9 @@ class Configuration(VssManager):
                 else:
                     # New configuration file. A new endpoint must be configured
                     f_type = 'Default template'
-                    config_endpoint = self._create_endpoint_config()
+                    config_endpoint = (
+                        new_endpoint or self._create_endpoint_config()
+                    )
                     config_file_tmpl.update_endpoint(config_endpoint)
                     # load and dump
                     config_file_dict = json.loads(config_file_tmpl.to_json())
@@ -690,7 +753,7 @@ class Configuration(VssManager):
                     self.write_config_file(new_endpoint=endpoint_cfg)
                 # profile exists
                 elif e_username and e_password and config_endpoint.token:
-                    username = e_username.decode('utf-8') if e_username else ''
+                    username = bytes_to_str(e_username) if e_username else ''
                     confirm = replace or click.confirm(
                         f"Would you like to replace existing configuration?\n "
                         f"{self.endpoint_name}:"
@@ -776,23 +839,25 @@ class Configuration(VssManager):
         return [objects[index]]
 
     def get_vskey_stor(self, **kwargs) -> bool:
-        """Create WebDav client to interact with remote CrushFTP."""
+        """Create s3 client to interact with remote minIO."""
         try:
-            from webdav3 import client as wc
+            from minio import Minio
         except ImportError:
             raise VssCliError(
-                'webdavclient3 dependency not found. '
+                'minio dependency not found. '
                 'try running "pip install vss-cli[stor]"'
             )
 
-        options = dict(
-            webdav_login=self.username,
-            webdav_password=self.password,
-            webdav_hostname=self.webdav_server,
-            verbose=self.verbose,
+        super().get_vskey_stor(
+            user=self.username,
+            password=self.password,
+            s3_endpoint=self.s3_server,
         )
-        self.vskey_stor = wc.Client(options=options)
-        return self.vskey_stor.valid()
+        _LOGGING.debug(
+            f's3_endpoint={self.vskey_stor_s3api} '
+            f'vskey_stor_s3_gui={self.vskey_stor_s3_gui}'
+        )
+        return self.vskey_stor
 
     def get_vm_by_id_or_name(self, vm_id: str, silent=False) -> Optional[List]:
         """Get virtual machine by identifier or name."""
@@ -845,6 +910,25 @@ class Configuration(VssManager):
                 return [v[index]]
             return v
 
+    def get_vm_snapshot_by_id_name_or_desc(
+        self, vm_id: str, id_name_or_desc: str
+    ) -> List[Dict]:
+        """Get vm snapshot by id name or description."""
+        snapshots = self.get_vm_snapshots(vm_id)
+        attributes = [('id', int), ('name', str), ('description', str)]
+        objs = self._filter_objects_by_attrs(
+            id_name_or_desc, snapshots, attributes
+        )
+        if not objs:
+            raise click.BadParameter(f'{id_name_or_desc} could not be found')
+        d_count = len(objs)
+        if d_count > 1:
+            return self.pick(
+                objs,
+                options=[f"{i['name']} ({i['description']})" for i in objs],
+            )
+        return objs
+
     def get_domain_by_name_or_moref(self, name_or_moref: str) -> List[Dict]:
         """Get domain by name or mo reference."""
         g_domains = self.get_domains()
@@ -1007,6 +1091,29 @@ class Configuration(VssManager):
             self.get_isos, name_or_path_or_id
         )
 
+    def get_clib_deployable_item_by_name_or_id_path(
+        self, name_or_id_or_path: Union[str, int]
+    ):
+        """Get content library deployable items."""
+        items = self.get_content_library_items(
+            show_all=True,
+            sort='path,desc',
+            per_page=500,
+            filter="type,in,OVF,VM_TEMPLATE",
+        )
+        attrs = [('id', int), ('name', str)]
+        objs = self._filter_objects_by_attrs(name_or_id_or_path, items, attrs)
+        # check if there's no ref
+        if not objs:
+            raise click.BadParameter(
+                f'{name_or_id_or_path} could not be found'
+            )
+        # count for dup results
+        o_count = len(objs)
+        if o_count > 1:
+            return self.pick(objs, options=[f"{i['name']}" for i in objs])
+        return objs
+
     def get_vm_image_by_name_or_id_path(
         self, name_or_path_or_id: Union[str, int]
     ) -> List[Any]:
@@ -1118,22 +1225,54 @@ class Configuration(VssManager):
     def get_api_spec_from_cli_spec(self, payload: dict, built: str) -> Dict:
         """Get API specification from CLI specification."""
         try:
+            from vss_cli.plugins.compute_plugins.callbacks import (
+                process_user_data,
+                process_day_zero,
+            )
+
             spec_payload = dict()
             # sections
             machine_section = payload['machine']
             networking_section = payload['networking']
             metadata_section = payload['metadata']
-            if built == 'os_install':
-                # machine section parse and update
+
+            if built in ['contentlib', 'os_install']:
+                # other
+                machine_section['power_on'] = machine_section.get(
+                    'power_on', False
+                )
                 spec_payload.update(machine_section)
-                # replace with valid values
+                spec_payload['disks'] = (
+                    [
+                        {
+                            "capacity_gb": d['capacity_gb'],
+                            "backing_mode": self.get_vm_disk_backing_mode_by_name(  # NOQA:
+                                d['backing_mode']
+                            )[
+                                0
+                            ][
+                                'type'
+                            ]
+                            if d.get('backing_mode')
+                            else 'persistent',
+                            "backing_sharing": self.get_vm_disk_backing_sharing_by_name(  # NOQA:
+                                d['backing_sharing']
+                            )[
+                                0
+                            ][
+                                'type'
+                            ]
+                            if d.get('backing_sharing')
+                            else 'sharingnone',
+                        }
+                        for d in machine_section['disks']
+                    ]
+                    if 'disks' in machine_section
+                    else []
+                )
                 spec_payload['os'] = self.get_os_by_name_or_guest(
                     machine_section['os']
                 )[0]['guest_id']
-                spec_payload['iso'] = self.get_iso_by_name_or_path(
-                    machine_section['iso']
-                )[0]['path']
-                # folder
                 spec_payload['folder'] = self.get_folder_by_name_or_moref_path(
                     machine_section['folder']
                 )[0]['moref']
@@ -1147,33 +1286,6 @@ class Configuration(VssManager):
                     }
                     for n in networking_section['interfaces']
                 ]
-                spec_payload['disks'] = [
-                    {
-                        "capacity_gb": d['capacity_gb'],
-                        "backing_mode": self.get_vm_disk_backing_mode_by_name(
-                            d['backing_mode']
-                        )[0]['type']
-                        if d.get('backing_mode')
-                        else 'persistent',
-                        "backing_sharing": self.get_vm_disk_backing_sharing_by_name(  # NOQA:
-                            d['backing_sharing']
-                        )[
-                            0
-                        ][
-                            'type'
-                        ]
-                        if d.get('backing_sharing')
-                        else 'sharingnone',
-                    }
-                    for d in machine_section['disks']
-                ]
-                # other
-                machine_section['high_io'] = machine_section.get(
-                    'high_io', False
-                )
-                machine_section['power_on'] = machine_section.get(
-                    'power_on', False
-                )
                 # metadata section
                 spec_payload.update(metadata_section)
                 spec_payload['built'] = built
@@ -1199,6 +1311,67 @@ class Configuration(VssManager):
                     spec_payload['admin'] = (
                         f"{admin_name}:" f"{admin_phone}:{admin_email}"
                     )
+
+            if built == 'contentlib':
+                spec_payload[
+                    'item_id'
+                ] = self.get_clib_deployable_item_by_name_or_id_path(
+                    machine_section['item']
+                )[
+                    0
+                ][
+                    'id'
+                ]
+                cloud_init = payload.get('cloud_init')
+                if cloud_init is not None:
+                    data, encoding = process_user_data(
+                        self,
+                        'user_data',
+                        cloud_init['user_data'],
+                    )
+                    udata_pload = {
+                        'userdata': data,
+                        'userdata_encoding': encoding,
+                    }
+                    if 'network_data' in cloud_init:
+                        ndata, ndata_encoding = process_user_data(
+                            self, 'network_data', cloud_init['network_data']
+                        )
+                        ndata_pload = {
+                            'networkconfig': ndata,
+                            'networkconfig_encoding': ndata_encoding,
+                        }
+                        udata_pload.update(ndata_pload)
+                    spec_payload['user_data'] = udata_pload
+                # additional parameters
+                add_params = payload.get('additional_parameters')
+                if add_params is not None:
+                    spec_payload[
+                        'additional_parameters'
+                    ] = validate_json_file_or_type(
+                        self, 'additional_parameters', add_params
+                    )
+                # day_zero
+                day_zero = payload.get('day-zero')
+                if day_zero is not None:
+                    _d0 = process_day_zero(
+                        self, 'day-zero', day_zero['config']
+                    )
+                    day_zero_payload = {
+                        'config': _d0[0],
+                        'config-encoding': _d0[1],
+                    }
+                    if 'id-token' in day_zero:
+                        _i0 = process_day_zero(
+                            self, 'id-token', day_zero['id-token']
+                        )
+                        _i0_d = {'idtoken': _i0[0], 'idtoken-encoding': _i0[1]}
+                        day_zero_payload.update(_i0_d)
+                    spec_payload['day_zero'] = day_zero_payload
+            if built == 'os_install':
+                spec_payload['iso'] = self.get_iso_by_name_or_path(
+                    machine_section['iso']
+                )[0]['path']
             return spec_payload
         except KeyError as ex:
             raise click.BadParameter(f'Invalid CLI specification: {ex}')
