@@ -5,14 +5,13 @@ import pkg_resources
 
 PACKAGE_NAME = "vss_cli"
 
-__version__ = "0.10.1"
+__version__ = "2022.6.0"
 
-REQUIRED_PYTHON_VER = (3, 6, 4)
 
 DEFAULT_TIMEOUT = 30
 DEFAULT_ENDPOINT = "https://cloud-api.eis.utoronto.ca"
 DEFAULT_ENDPOINT_NAME = "cloud-api"
-DEFAULT_WEBDAV_SERVER = "https://vskey-stor.eis.utoronto.ca"
+DEFAULT_S3_SERVER = "https://vskey-stor.eis.utoronto.ca"
 _LEGACY_CONFIG = ("~", ".vss-cli", "config.json")
 _DEFAULT_CONFIG = ("~", ".vss-cli", "config.yaml")
 _DEFAULT_HISTORY = ("~", ".vss-cli", "history")
@@ -27,6 +26,7 @@ DEFAULT_DATA_PATH = pkg_resources.resource_filename(PACKAGE_NAME, "data")
 DEFAULT_CONFIG_TMPL = os.path.join(DEFAULT_DATA_PATH, "config.yaml")
 DEFAULT_CHECK_UPDATES = True
 DEFAULT_CHECK_MESSAGES = True
+DEFAULT_TOTP = False
 
 DEFAULT_TABLE_FORMAT = "simple"
 DEFAULT_DATA_OUTPUT = "table"
@@ -34,10 +34,13 @@ DEFAULT_RAW_OUTPUT = "json"
 DEFAULT_OUTPUT = "auto"
 DEFAULT_VERBOSE = False
 DEFAULT_DEBUG = False
+DEFAULT_COLUMNS_WIDTH = -1
+DEFAULT_WAIT_FOR_REQUESTS = False
 
 DEFAULT_SETTINGS = {
     "endpoint": DEFAULT_ENDPOINT,
     "output": DEFAULT_OUTPUT,
+    "s3_server": DEFAULT_S3_SERVER,
     "table_format": DEFAULT_TABLE_FORMAT,
     "check_for_messages": DEFAULT_CHECK_MESSAGES,
     "check_for_updates": DEFAULT_CHECK_UPDATES,
@@ -55,6 +58,7 @@ GENERAL_SETTINGS = {
     "verbose": bool,
     "default_endpoint_name": str,
     "output": str,
+    "s3_server": str,
     "table_format": str,
     "timeout": int,
     "columns_width": int,
@@ -100,6 +104,7 @@ DEFAULT_VM_DEL_MSG = (
 COLUMNS_TWO_FMT = "{0:<20}: {1:<20}"
 
 COLUMNS_DEFAULT = [("all", "*")]
+COLUMNS_MFA_MIN = [("message",), ("type",)]
 COLUMNS_VM_MIN = [("moref",), ("name",)]
 COLUMNS_VIM_REQUEST = [("vm_moref",), ("vm_name",)]
 COLUMNS_MOREF = [("moref",), ("name",)]
@@ -124,6 +129,7 @@ COLUMNS_PERMISSION = [("principal",), ("group",), ("propagate",)]
 COLUMNS_MIN = [("id",), ("created_on",), ("updated_on",)]
 COLUMNS_VSS_SERVICE = [("id",), ("label",), ("name",), ("group.name",)]
 COLUMNS_IMAGE = [("id",), ("path",), ("name",)]
+COLUMNS_CLIB_ITEMS = [("id",), ("name",), ("library.name",), ('size',)]
 COLUMNS_OS = [("id",), ("guest_id",), ("full_name",), ("family",)]
 COLUMNS_REQUEST = [*COLUMNS_MIN, ("status",)]
 COLUMNS_REQUEST_WAIT = [('warnings', 'warnings[*]'), ('errors', 'errors[*]')]
@@ -133,12 +139,20 @@ COLUMNS_REQUEST_MAX = [
     ("task_id", "task_id"),
     ("user.username",),
 ]
-COLUMNS_REQUEST_IMAGE_SYNC_MIN = [*COLUMNS_REQUEST, ("type",)]
-COLUMNS_REQUEST_IMAGE_SYNC = [
-    *COLUMNS_REQUEST,
-    ("type",),
+COLUMNS_REQUEST_FILE_SYNC = [
     ("deleted",),
     ("added",),
+]
+COLUMNS_REQUEST_VMDK_SYNC_MIN = [*COLUMNS_REQUEST, ('run_time',)]
+COLUMNS_REQUEST_VMDK_SYNC = [
+    *COLUMNS_REQUEST_VMDK_SYNC_MIN,
+    *COLUMNS_REQUEST_FILE_SYNC,
+    *COLUMNS_REQUEST_MAX,
+]
+COLUMNS_REQUEST_IMAGE_SYNC_MIN = [*COLUMNS_REQUEST, ("type",)]
+COLUMNS_REQUEST_IMAGE_SYNC = [
+    *COLUMNS_REQUEST_IMAGE_SYNC_MIN,
+    *COLUMNS_REQUEST_FILE_SYNC,
     *COLUMNS_REQUEST_MAX,
 ]
 COLUMNS_REQUEST_SUBMITTED = [
@@ -153,25 +167,51 @@ COLUMNS_REQUEST_MULT_SUBMITTED = [
     ("task_id", "request.task_id[*]"),
     ("message",),
 ]
+COLUMNS_REQUEST_RETIRE = [
+    *COLUMNS_REQUEST,
+    *COLUMNS_VIM_REQUEST,
+    ("retire_on",),
+    ("warning",),
+    ("warning_on",),
+]
+COLUMNS_REQUEST_RETIRE_CANCEL = [
+    *COLUMNS_REQUEST_RETIRE,
+    ("cancelled_on",),
+    ("cancelled_by_user.username",),
+]
+COLUMNS_REQUEST_RETIRE_CONFIRM = [
+    *COLUMNS_REQUEST_RETIRE,
+    ("confirmed_on",),
+    ("confirmed_by_user.username",),
+]
 COLUMNS_REQUEST_SNAP_MIN = [
     *COLUMNS_REQUEST,
     *COLUMNS_VIM_REQUEST,
     ("action",),
-    ("snapshot.from_date",),
-    ("snapshot.to_date",),
+    ("from_date",),
+    ("to_date",),
 ]
 COLUMNS_REQUEST_SNAP = [
     *COLUMNS_REQUEST_SNAP_MIN,
-    ("snapshot.description",),
-    ("snapshot.snap_id",),
+    ("description",),
+    ("snap_id",),
     ("extensions",),
 ]
 COLUMNS_REQUEST_CHANGE_MIN = [
     *COLUMNS_REQUEST,
     *COLUMNS_VIM_REQUEST,
-    ("approval.approved",),
+    ("approved",),
     ("attribute",),
 ]
+COLUMNS_REQUEST_CHANGE_MIN_VM = [
+    ("id",),
+    ("created_on",),
+    ("status",),
+    *COLUMNS_VIM_REQUEST,
+    ("user.username",),
+    ("attribute",),
+    ("value[*]",),
+]
 COLUMNS_REQUEST_CHANGE = [
     *COLUMNS_REQUEST_CHANGE_MIN,
     ("value", "value[*]"),
@@ -207,16 +247,15 @@ COLUMNS_REQUEST_INVENTORY = [
 COLUMNS_REQUEST_NEW_MIN = [
     *COLUMNS_REQUEST,
     *COLUMNS_VIM_REQUEST,
-    ("approval.approved",),
+    ("approved",),
     ("built_from",),
 ]
 COLUMNS_REQUEST_NEW = [
     *COLUMNS_REQUEST_NEW_MIN,
     ('guest_os',),
     ("domain",),
-    ("source_vm",),
-    ("source_template",),
-    ("source_image",),
+    ("source",),
+    ("retirement",),
     ("folder",),
     ("cpu",),
     ("memory",),
@@ -341,14 +380,16 @@ COLUMNS_VM_CTRL = [
     ("controller_key",),
     ("summary",),
     ("shared_bus",),
-    ("hot_add_rmove",),
+    ("hot_add_remove",),
 ]
 COLUMNS_VM_DISK_MIN = [
     ("label",),
     ("unit",),
-    ("controller.virtual_device_node",),
+    ("controller.label",),
+    ("capacity_gib",),
+    ("notes",),
 ]
-COLUMNS_VM_DISK = [*COLUMNS_VM_DISK_MIN, ("capacity_gb",), ("shares.level",)]
+COLUMNS_VM_DISK = [*COLUMNS_VM_DISK_MIN, ("shares.level",)]
 
 COLUMNS_VM_DISK_BACKING = [
     ("descriptor_file_name",),
@@ -382,6 +423,7 @@ COLUMNS_VM_STATE = [
     ("domain_moref", "domain.moref"),
 ]
 COLUMNS_VM_TOOLS = [("version",), ("version_status",), ("running_status",)]
+COLUMNS_VM_TPM = [('label',), ('summary',), ('key',)]
 COLUMNS_VM_HW = [
     ("value",),
     ("status",),
@@ -394,10 +436,10 @@ COLUMNS_VSS_OPTIONS = [("options", "[*]")]
 COLUMNS_GROUP = [
     ("name",),
     ("description",),
-    ("members.count",),
+    ("users_count",),
     ("ldap.last_sync",),
 ]
-COLUMNS_GROUPS = [("id",), ("name",), ("description",), ('members.count',)]
+COLUMNS_GROUPS = [("id",), ("name",), ("description",), ('users_count',)]
 COLUMNS_GROUP_MEMBERS = [
     ("username",),
     ("first_name",),
@@ -424,6 +466,12 @@ COLUMNS_USER_STATUS = [
     ("last_access",),
     ("ip_address",),
 ]
+COLUMNS_USER_MFA = [
+    ("enabled",),
+    ("method",),
+    ("enabled_on",),
+    ("disabled_on",),
+]
 COLUMNS_MESSAGE_DIGEST = [("message",)]
 COLUMNS_NOT_REQUEST = [
     ("all",),
@@ -432,8 +480,14 @@ COLUMNS_NOT_REQUEST = [
     ("error",),
     ("submission",),
 ]
-COLUMNS_WEBDAV = [("files", "[*]")]
-COLUMNS_WEBDAV_INFO = [("created",), ("modified",), ("name",), ("size",)]
+COLUMNS_STOR = [("files", "[*]")]
+COLUMNS_STOR_INFO = [
+    ("name",),
+    ("bucket_name",),
+    ("last_modified",),
+    ("size",),
+]
+COLUMNS_STOR_SHARE = [('url',)]
 COLUMNS_SSH_KEY_MIN = [*COLUMNS_MIN, ("type",), ("comment",)]
 COLUMNS_SSH_KEY = [*COLUMNS_SSH_KEY_MIN, ("fingerprint",), ("value",)]
 COLUMNS_VMRC = [("enabled",), ("options",)]
